#include "ray.h"
#include <qmessagebox.h>

using namespace qglviewer;

Ray::Ray()
{
  setStart(Vec());
  setDirection(Vec(0.0, 0.0, 1.0));
}

Ray::Ray(const Vec start, const Vec direction)
{
  setStart(start);
  setDirection(direction);
}

void Ray::setDirection(const Vec& direction)
{
  if (direction.norm() < 1e-8)
    {
      QMessageBox::critical(NULL, "Ray::setDirection error", "Null length direction in Ray::setDirection");
      return;
    }
  direction_ = direction;
  direction_.normalize();

  ////////
  inv_direction = Vec(1/direction.x, 1/direction.y, 1/direction.z);
  sign[0] = (inv_direction.x < 0);
  sign[1] = (inv_direction.y < 0);
  sign[2] = (inv_direction.z < 0);
}

// Draw a ray
void Ray::draw() const
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  glPushAttrib(GL_ALL_ATTRIB_BITS);

  glDisable(GL_LIGHTING);

  glColor3f(0,0,1);

  glBegin(GL_LINES);
    glVertex3f(start().x, start().y, start().z);
    Vec end=start()+direction()*10;
    glVertex3f(end.x, end.y, end.z);
  glEnd();

  glPopAttrib();
  glPopMatrix();
}
