#ifndef RAY_TRACER_H
#define RAY_TRACER_H

#include <qimage.h>
#include "color.h"
#include "ray.h"
#include "hit.h"

class Scene;

class RayTracer
{
public :
  RayTracer() : scene_(NULL), maxReboundCount(5), imgTexID(0) {};

  const Scene* const scene() const { return scene_; }
  const QImage& image() const { return image_; }

  void setScene(const Scene* const scene) { scene_ = scene; }

  // "optLight": 0 -> Phong, 1 -> Blinn-Phong
  // "optRefrac": 0 -> without refraction, 1 -> with refraction
  void renderImage() const;
  void saveImage(const QString& name, bool overwrite=false) const;


	Color rayColor(const Ray& ray, int numReb=0, qglviewer::Vec oinray=qglviewer::Vec(0,0,0)) const;
	Color rayColorWithRefraction(const Ray& ray, int reflectDepth=0, int refractDepth=0, float prevRIndex=1.0f) const;


  void drawRayPath();
  void drawImage(float size=4);

private:
  const Scene* scene_;
  mutable QImage image_;

  int maxReboundCount;
  mutable GLuint imgTexID;

  class Segment {
	public:
		Segment(){}
		Segment(const Ray &ray, const Hit &hit){
			start=ray.start();
			end=hit.intersection();
			endNormal=hit.normal();
		}
		//No-intersection segment
		Segment(const Ray &ray){
			start=ray.start();
			end=ray.start()+ray.direction()*1000;
			endNormal=qglviewer::Vec(0,0,0);
		}

		qglviewer::Vec start;
		qglviewer::Vec end;
		qglviewer::Vec endNormal;
	};

  mutable QVector<Segment> rayPath_;

};

#endif // RAY_TRACER_H
