#include "triangle.h"
#include "material.h"
#include "RTConfig.h"

// Pour ne pas avoir a mettre qglviewer:: devant les classes
// definies dans le namespace qglviewer.
using namespace qglviewer;

Triangle::Triangle() {
}

void Triangle::draw() const
{
	if(!RTConfig::getInstance()->drawTriangle)
		return;

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  glPushAttrib(GL_ALL_ATTRIB_BITS);

  glMultMatrixd(frame().matrix());
  glColor3fv(material().diffuseColor());

  float colorMat[] = { material().diffuseColor().r, material().diffuseColor().g, material().diffuseColor().b, 1.0f };
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colorMat);
  float colorMatSpec[] = { material().specularColor().r, material().specularColor().g, material().specularColor().b, 1.0f };
  glMaterialfv(GL_FRONT, GL_SPECULAR, colorMatSpec);

  glMateriali(GL_FRONT, GL_SHININESS, (int)material().specularCoefficient());

  glBegin(GL_TRIANGLES);
  glVertex3f(p0.x, p0.y, p0.z);
  glVertex3f(p1.x, p1.y, p1.z);
  glVertex3f(p2.x, p2.y, p2.z);
  glEnd();

  glPopAttrib();
  glPopMatrix();
}

void Triangle::initFromDOMElement(const QDomElement& e)
{
  // A FAIRE
  Object::initFromDOMElement(e);

  QDomElement child=e.firstChild().toElement();
  while (!child.isNull()){
      if (child.tagName() == "p0")
		this->p0=Vec(child);
      if (child.tagName() == "p1")
		this->p1=Vec(child);
	  if (child.tagName() == "p2")
		this->p2=Vec(child);

      child = child.nextSibling().toElement();
  }
}

bool Triangle::intersect(const Ray& ray, Hit& hit) const
{
  float t, u, v;
  
  Vec dir=ray.direction();
  Vec orig=ray.start();

  float EPSILON=0.0001;

  Vec vert0=this->frame().inverseCoordinatesOf(p0);
  Vec vert1=this->frame().inverseCoordinatesOf(p1);
  Vec vert2=this->frame().inverseCoordinatesOf(p2);

  ////Ray triangle intersection from "Fast Minimum Storage RayTriangle Intersection" [MollerTrumbore97]

   //double edge1[3], edge2[3], tvec[3], pvec[3], qvec[3];
   Vec edge1, edge2, pvec, tvec, qvec;
   float det,inv_det;

   /* find vectors for two edges sharing vert0 */
   edge1= vert1-vert0;
   edge2= vert2-vert0;

   /* begin calculating determinant - also used to calculate U parameter */
   pvec= cross(dir, edge2);

   /* if determinant is near zero, ray lies in plane of triangle */
   det = edge1 * pvec; //dot

   if (det > -EPSILON && det < EPSILON)
     return false;

   inv_det = 1.0f / det;

   /* calculate distance from vert0 to ray origin */
   tvec = orig - vert0;

   /* calculate U parameter and test bounds */
   u = (tvec * pvec) * inv_det; //dot
   if (u < 0.0 || u > 1.0)
     return 0;

   /* prepare to test V parameter */
   qvec= cross(tvec, edge1);

   /* calculate V parameter and test bounds */
   v = (dir * qvec) * inv_det; //dot
   if (v < 0.0 || u + v > 1.0)
     return 0;

   /* calculate t, ray intersects triangle */
   t = (edge2 * qvec) * inv_det; //dot


  float tmin=RTConfig::getInstance()->getRayTMin();// 0.0001;

  if(hit.time()>t && t>tmin){
	  hit.setTime(t);
	  hit.setIntersection( ray.direction()*t+ray.start() );
	  hit.setMaterial(this->material());
	  hit.setNormal(cross(edge1, edge2));

	  // "Partie XI: Textures"
	  hit.setUV(u, v );

	  return true;
  }

  return false;
}
