#include "viewer.h"
#include "scene.h"
#include <qfiledialog.h>
#include <math.h>

#include "sphere.h"
#include "plane.h"
#include "cubeMap.h"
#include "RTConfig.h"

using namespace std;
using namespace qglviewer;

Viewer::~Viewer()
{
  delete scene_;
}

void Viewer::draw()
{
	//glPushAttrib(GL_ALL_ATTRIB_BITS);

	//glDisable(GL_STENCIL_TEST);
	//glDisable(GL_ALPHA_TEST);

	/*glEnable(GL_STENCIL_TEST);
	glStencilFunc( GL_ALWAYS, 0, 1);
	glStencilOp( GL_KEEP, GL_KEEP, GL_INCR);*/



 // ray_.draw();
  //scene()->camera().testAllRays();
 // scene()->camera().drawAllRays();


	//std::cout<<Quaternion(Vec(0,0,1), 2*M_PI).q[0]<<" "<<Quaternion(Vec(0,0,1), 2*M_PI).q[1]<<" "<<Quaternion(Vec(0,0,1), 2*M_PI).q[2]<<" "<<Quaternion(Vec(0,0,1), 2*M_PI).q[3]<<"\n";


if(RTConfig::getInstance()->useGPURT){
#ifdef USE_GPURT

  //RTConfig::getInstance()->gpurtSafeMode=true;

  //std::cout<<this->camera()->zNear()<<" "<<this->camera()->zFar()<<"\n";

  glPushAttrib(GL_ALL_ATTRIB_BITS);

  //glEnable(GL_BLEND);
  //glDisable(GL_ALPHA_TEST);
  glDisable(GL_DEPTH_TEST);
  glDepthMask(false);
 // glDisable(GL_DEPTH_TEST);

  //glStencilFunc( GL_NOTEQUAL, 0, 1);
  //glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP);

  //glAlphaFunc(GL_GREATER, 0);

  if(this->gpuRT_){
	  //Lights update

	int plCount=0;
	int dlCount=0;
	bool needUpdate=false;
	for(int i=0; i<this->scene()->numLight(); i++){
		if( dynamic_cast<PointLight *>(this->scene()->getLight(i))){
			//New interface
			if(i==this->curManipulatedLight){
				this->gpuRT_->setPointLightPos(plCount, NemoGraphics::Vector3f(	this->scene()->getLight(i)->frame().position().v_) );
				this->gpuRT_->updateLights();
			}
			plCount++;
		}else if( dynamic_cast<DirectionalLight *>(this->scene()->getLight(i)) ){
			//this->gpuRT_->setPointLightPos(dlCount, NemoGraphics::Vector3f(	this->scene()->getLight(i)->frame().position().v_) );
			//this->gpuRT_->setDirLightOrient(dlCount, NemoGraphics::Vector3f(this->scene()->getLight(i)->frame().orientation().rotate(Vec ().v_) );

			dlCount++;
		}
	}
	

  ///////
	Mat4 tmpm;
	for(int i=0; i<4*4; i++)
	tmpm.mat[i]=this->camera()->frame()->/*inverse().*/matrix()[i];


	this->gpuRT_->viewMat=tmpm;//.transpose();

	  //this->gpuRT_->getViewPos()=Vector3f(this->camera()->frame()->position().v_);
	 // this->gpuRT_->getViewOrient()=Vector3f(this->camera()->frame()-> inverseTransformOf(Vec(1,1,1)).v_);



	this->gpuRT_->render();


  }

   glDisable(GL_DEPTH_TEST);
   glDepthMask(true);
   glEnable(GL_DEPTH_TEST);
   glColorMask(false, false, false, false);
   RTConfig::getInstance()->drawPlane=true;
   RTConfig::getInstance()->drawSphere=true;
   RTConfig::getInstance()->drawTriangle=false;
   RTConfig::getInstance()->drawTriangleMesh=false;
   RTConfig::getInstance()->drawLight=false;

   RTConfig::getInstance()->drawCubeMap=false;
   scene()->draw();

   glColorMask(true, true, true, true);
   glEnable(GL_DEPTH_TEST);
   glDepthMask(true);
   RTConfig::getInstance()->drawPlane=false;
   RTConfig::getInstance()->drawSphere=false;
   RTConfig::getInstance()->drawTriangle=true;
   RTConfig::getInstance()->drawTriangleMesh=true;
   RTConfig::getInstance()->drawLight=true;
   scene()->draw();



  glPopAttrib();

#endif
}else{
	//RTConfig::getInstance()->gpurtSafeMode=false;

   RTConfig::getInstance()->drawPlane=true;
   RTConfig::getInstance()->drawSphere=true;
   RTConfig::getInstance()->drawTriangle=true;
   RTConfig::getInstance()->drawTriangleMesh=true;
   RTConfig::getInstance()->drawCubeMap=true;

	glEnable(GL_DEPTH_TEST);
	scene()->draw();
}

  glEnable(GL_DEPTH_TEST);

  if(RTConfig::getInstance()->drawHelpers){
	 // Put the camera according to the radius of the scene
	  scene()->camera().draw(0.1);
	  
	  rayhit_.draw();
	  rayTracer_.drawRayPath();
  }

 
  rayTracer_.drawImage(RTConfig::getInstance()->previewSize);

}

void Viewer::init()
{

  // Key description for help window (press 'H')
	setKeyDescription(Qt::Key_L, "Loads a new scene");
	setKeyDescription(Qt::Key_D, "Ray Trace from scene camera view");
	setKeyDescription(Qt::SHIFT+Qt::Key_D, "Ray Trace from current view");
	setKeyDescription(Qt::Key_S, "Save ray tracing result");
	setKeyDescription(Qt::SHIFT+Qt::Key_S, "Shoot rays from <i>current</i> view point");
	setKeyDescription(Qt::Key_N, "Change number of rays");
	setKeyDescription(Qt::Key_V, "Debug Display");
	setKeyDescription(Qt::Key_I, "Switch Use of GPU RayTracing (if avalaible)");
	setKeyDescription(Qt::Key_1, "Change Blinn/Phong lighting mode");
	setKeyDescription(Qt::Key_2, "Switch Texturing");
	setKeyDescription(Qt::Key_W, "Increase max reflection depth");
	setKeyDescription(Qt::SHIFT+Qt::Key_W, "Decrease max reflection depth");
	setKeyDescription(Qt::Key_X, "Increase max refraction depth");
	setKeyDescription(Qt::SHIFT+Qt::Key_X, "Decrease max refraction depth");
	setKeyDescription(Qt::Key_P, "Switch preview mode (corner/camera)");
	setKeyDescription(Qt::Key_F, "Switch use fresnel");
	setKeyDescription(Qt::Key_O, "Switch test flag (for debugging purpose)");
	setKeyDescription(Qt::Key_3, "Switch current manipulated light");
	setKeyDescription(Qt::Key_6, "Start/Stop Annimation");
	setKeyDescription(Qt::Key_4, "Decrease corner preview size");
	setKeyDescription(Qt::Key_5, "Increase corner preview size");
	setKeyDescription(Qt::Key_7, "Switch helpers (lights, rays, camera) drawing");


  setHandlerKeyboardModifiers(QGLViewer::FRAME,  Qt::ControlModifier);
  setHandlerKeyboardModifiers(QGLViewer::CAMERA, Qt::NoModifier);

  setScene(new Scene());
  rayTracer_.setScene(scene());

  // So that transparent materials are correctly displayed.
  // Disable if rendering is too slow.
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

 

  // Restore previous viewer state (camera, key frames...)
  restoreStateFromFile();

  // Loads scene (prevents from pressing 'L' at each start).
  //loadScene("troisSpheres.scn");
  loadScene("troisSpheres.scn");
  
  // Set Camera to scene Camera. Set scene center and radius.
  initFromScene();

}

QString Viewer::helpString() const
{
  QString text("<h2>L a n c e r   d e   r a y o n s</h2>");
  text += "Ajoutez ici toute aide utile  l'utilisateur. ";
  text += "N'oubliez pas de commenter vos raccourcis clavier avec setKeyDescription().<br>";
  text += "Appuyez sur <b>Escape</b> pour quitter.";
  return text;
}

void Viewer::keyPressEvent(QKeyEvent *e)
{
  switch (e->key())
    {

	// Change the number of rays per pixel for rendering
	case Qt::Key_N :
	  if (e->modifiers() == Qt::ShiftModifier)
	  {
		  RTConfig::getInstance()->setNumPrimaryRays(RTConfig::getInstance()->getNumPrimaryRays()-1);
	  }else{
		  RTConfig::getInstance()->setNumPrimaryRays(RTConfig::getInstance()->getNumPrimaryRays()+1);
		  
	  }
	  print("Nombre de rayons: ", RTConfig::getInstance()->getNumPrimaryRays());
	  break;


    case Qt::Key_L :
      loadScene(QFileDialog::getOpenFileName(this, "Select a scene", ".", "Scenes (*.scn);;All files (*)"));
      break;


    // Save the image, option: Phong & without refraction
    case Qt::Key_S :	  
	{ //rayTracer().renderImage();
	  // Remplacer cette ligne par : const QString name = "result.jpg";
	  // pour ne pas avoir  choisir un nom  chaque fois.
	  const QString name = QFileDialog::getSaveFileName();
	  rayTracer().saveImage(name);

	  break;
	}

	
	case Qt::Key_V :	  
	{ //rayTracer().renderImage();
	  RTConfig::getInstance()->debugDisplay=!RTConfig::getInstance()->debugDisplay;
	  print("Debug Display : ", RTConfig::getInstance()->debugDisplay);

	  break;
	}

	case Qt::Key_I :	  
	{ //rayTracer().renderImage();
	  RTConfig::getInstance()->useGPURT=!RTConfig::getInstance()->useGPURT;
	  print("Use GPURT: ", RTConfig::getInstance()->useGPURT);
	  break;
	}

	// option: LightMode : Blinn-Phong or Phong
    case Qt::Key_1 :
		RTConfig::getInstance()->setLightMode(!RTConfig::getInstance()->getLightMode());
		print("Light Mode: ",RTConfig::getInstance()->getLightMode());
		break;
	case Qt::Key_2 :
		RTConfig::getInstance()->setTexEnable(!RTConfig::getInstance()->getTexEnable());
		print("Texturing : ",RTConfig::getInstance()->getTexEnable());
		break;

	case Qt::Key_W :
	  if (e->modifiers() == Qt::ShiftModifier)
	  {
			RTConfig::getInstance()->setMaxReflection( RTConfig::getInstance()->getMaxReflection()-1);
	  }else{
			RTConfig::getInstance()->setMaxReflection( RTConfig::getInstance()->getMaxReflection()+1);
		  
	  }
	  print("Max reflection: ",RTConfig::getInstance()->getMaxReflection());
	  break;


	case Qt::Key_X :
	  if (e->modifiers() == Qt::ShiftModifier)
	  {
			RTConfig::getInstance()->setMaxRefraction( RTConfig::getInstance()->getMaxRefraction()-1);
	  }else{
			RTConfig::getInstance()->setMaxRefraction( RTConfig::getInstance()->getMaxRefraction()+1);
		  
	  }
	  print("Max refraction: ", RTConfig::getInstance()->getMaxRefraction());
	  break;


    case Qt::Key_P :
		RTConfig::getInstance()->setPreviewMode(!RTConfig::getInstance()->getPreviewMode());

		print("Preview Mode: ",RTConfig::getInstance()->getPreviewMode());
	  break;

    case Qt::Key_F :
		RTConfig::getInstance()->useFresnel=!RTConfig::getInstance()->useFresnel;

		print("USe Fresnel Mode: ", RTConfig::getInstance()->useFresnel);
	  break;

	case Qt::Key_0 :
		RTConfig::getInstance()->testFlag=!RTConfig::getInstance()->testFlag;
		print("Test Flag: ",RTConfig::getInstance()->testFlag);
		break;
	case Qt::Key_3 :
		curManipulatedLight=(curManipulatedLight+1)%this->scene()->numLight();
		print("Current light: ",curManipulatedLight);

		setManipulatedFrame(&(this->scene()->getLight(curManipulatedLight)->manipulatedFrame()));
		break;


	case Qt::Key_6 :
		if(this->animationIsStarted())
			this->stopAnimation();
		else
			this->startAnimation();

		print("Animation ", this->animationIsStarted());
		break;

	case Qt::Key_4 :
		RTConfig::getInstance()->previewSize-=1.0f;

		print("Preview Size ", RTConfig::getInstance()->previewSize);
		break;
	case Qt::Key_5 :
		RTConfig::getInstance()->previewSize+=1.0f;

		print("Preview Size ", RTConfig::getInstance()->previewSize);
		break;
	case Qt::Key_7 :
		RTConfig::getInstance()->drawHelpers=!RTConfig::getInstance()->drawHelpers;

		print("Draw helpers ", RTConfig::getInstance()->drawHelpers);
		break;

	case Qt::Key_D :
      if ((e->modifiers() == Qt::NoModifier) || (e->modifiers() == Qt::ShiftModifier))	{
	  if (e->modifiers() == Qt::ShiftModifier)  {
	      // Shift+S renders image from current view point
	      scene()->camera().frame().setPosition(camera()->position());
	      scene()->camera().frame().setOrientation(camera()->orientation());
		  //scene()->camera().setFieldOfView(camera()->fieldOfView());
	    }
	  
	  rayTracer().renderImage();
	  

	  if ((e->modifiers() == Qt::ShiftModifier) && (camera()->keyFrameInterpolator(1))) {
	      // Restore initial scene camera
	      scene()->camera().frame().setPosition(camera()->keyFrameInterpolator(1)->keyFrame(0).position());
	      scene()->camera().frame().setOrientation(camera()->keyFrameInterpolator(1)->keyFrame(0).orientation());
	    }
	  break;
	}


    default :
      QGLViewer::keyPressEvent(e);
    }

	this->updateGL();
}

void Viewer::loadScene(const QString& name)
{
  if (name.isEmpty())
    return;

  this->curManipulatedLight=0;
  setManipulatedFrame( camera()->frame() );


  scene()->loadFromFile(name);
  
  // Change QGLViewer settings according to scene
  setSceneCenter(scene()->center());
  setSceneRadius(scene()->radius(sceneCenter()));

  //setSceneRadius(5);



  // Set first camera key frame (F1) to camera position.
  initFromScene();

#ifdef USE_GPURT
  initGPURT();
  RTConfig::getInstance()->useGPURT=true;
#endif
}

// Make the first keyFrameInterpolator (binded to F1) contain one position, that cooresponds to the
// scene's camera position. Attention, camera() is moved to camera position.
void Viewer::initFromScene()
{
  camera()->setPosition(scene()->camera().frame().position());
  camera()->setOrientation(scene()->camera().frame().orientation());
  camera()->setFieldOfView(scene()->camera().fieldOfView());
  
  // Remove previous keyFrames in path 1 (if any)
  if (camera()->keyFrameInterpolator(1))
    camera()->keyFrameInterpolator(1)->deletePath();

  // Add current (i.e. scene camera) position to F1.
  camera()->addKeyFrameToPath(1);
}


// Get the point where the mouse clicked
void Viewer::select(const QPoint& point)
{
  Vec orig, dir;

  camera()->convertClickToLine(point,orig, dir);
  ray_.setStart(orig);
  ray_.setDirection(dir);
  //this->scene()->intersect(ray_, rayhit_);  // replaced by "rayTracer_.rayColor(ray_);"

  // "Partie X: Plusieurs Rebonds"
  rayhit_.setTime(1.0E5);
  rayhit_.setIntersection(Vec(0,0,0));
  //rayTracer_.rayColor(ray_);
  rayTracer_.rayColorWithRefraction(ray_);


}

#ifdef USE_GPURT
void Viewer::initGPURT() {
	if(gpuRT_)
		delete gpuRT_;

	gpuRT_=new GPURT(NemoGraphics::Vector2i(this->scene()->camera().xResolution(), this->scene()->camera().yResolution()) );

	gpuRT_->setFOV(this->scene()->camera().fieldOfView());

	if(this->scene()->specialMode==10)
		this->gpuRT_->init(1);
	else
		this->gpuRT_->init(0);

    gpuRT_->setNearFarClipPlanes(this->camera()->zNear(), this->camera()->zFar());


	gpuRT_->setSpecialMode(this->scene()->specialMode);
	
	Sphere *sp;
	Plane *plane;
	CubeMap *cubeMap;

	int curtex=-1, curbump=-1;

	//Point lights managment
	PointLight *pl;
	std::vector<PointLight *> plv;
	//Dir lights
	DirectionalLight *dl;
	AmbientLight *al;
	std::vector<DirectionalLight *> dlv;

	int npl, ndl;

	for(int i=0; i<this->scene()->numLight(); i++){
		if( pl=dynamic_cast<PointLight *>(this->scene()->getLight(i)) ){
			plv.push_back(pl);
			//New interface
			this->gpuRT_->addPointLight(Vector3f(pl->color().r, pl->color().g, pl->color().b), Vector3f(pl->position().v_));
		}else if( dl=dynamic_cast<DirectionalLight *>(this->scene()->getLight(i)) ){
			dlv.push_back(dl);
			this->gpuRT_->addDirLight(Vector3f(dl->color().r, dl->color().g, dl->color().b), Vector3f(dl->direction().v_));
		}else if( al=dynamic_cast<AmbientLight *>(this->scene()->getLight(i)) ){
			this->gpuRT_->setAmbientLight(Vector3f(al->color().r, al->color().g, al->color().b));
		}
	}


	//Ajout
	/*this->gpuRT_->setNumLights(plv.size(), dlv.size());
	for(int i=0; i<plv.size(); i++){
		this->gpuRT_->setPointLight(i, Vector3f(plv[i]->color().r, plv[i]->color().g, plv[i]->color().b),
			Vector3f(plv[i]->position().v_) );

	}
	for(int i=0; i<dlv.size(); i++){
		this->gpuRT_->setDirLight(i, Vector3f(dlv[i]->color().r, dlv[i]->color().g, dlv[i]->color().b),
			Vector3f(dlv[i]->direction().v_) );

	}*/

	////////////////

	
	for(int i=0; i<this->scene()->numObject(); i++){
		if(sp=dynamic_cast<Sphere *>(this->scene()->getObject(i))){

			if(!sp->material().texture().isNull()){
				curtex=this->gpuRT_->addTexture(
					Vector3i(sp->material().texture().size().width(), sp->material().texture().size().height(), 0), 
					GL_BGRA, GL_UNSIGNED_BYTE, sp->material().texture().bits());
			}else
				curtex=-1;

			if(!sp->material().bumpMapTexture().isNull()){
				curbump=this->gpuRT_->addBump(
					Vector3i(sp->material().bumpMapTexture().width(), sp->material().bumpMapTexture().height(), 0), 
					GL_BGRA, GL_UNSIGNED_BYTE, sp->material().bumpMapTexture().bits());
			}else
				curbump=-1;


			this->gpuRT_->addSphere(Vector3f(sp->frame().position().v_), sp->radius(), 
				Vector3f(sp->material().diffuseColor().r, sp->material().diffuseColor().g, sp->material().diffuseColor().b),
				Vector3f(sp->material().specularColor().r, sp->material().specularColor().g, sp->material().specularColor().b) , 
				sp->material().specularCoefficient(), 
				curtex, curbump, 
				Vector2f(1.0f/sp->material().textureScaleU(), 1.0f/sp->material().textureScaleV()),
				Vector3f(sp->material().reflectiveColor().r, sp->material().reflectiveColor().g, sp->material().reflectiveColor().b),
				Vector3f(sp->material().refractiveColor().r, sp->material().refractiveColor().g, sp->material().refractiveColor().b),
				1.0f/sp->material().refractiveIndex());

		}else if(plane=dynamic_cast<Plane *>(this->scene()->getObject(i))){

			if(!plane->material().texture().isNull()){
				curtex=this->gpuRT_->addTexture(
					Vector3i(plane->material().texture().size().width(), plane->material().texture().size().height(), 0), 
					GL_BGRA, GL_UNSIGNED_BYTE, plane->material().texture().bits());

			}else
				curtex=-1;

			if(!plane->material().bumpMapTexture().isNull()){
				curbump=this->gpuRT_->addBump(
					Vector3i(plane->material().bumpMapTexture().size().width(), plane->material().bumpMapTexture().size().height(), 0), 
					GL_BGRA, GL_UNSIGNED_BYTE, plane->material().bumpMapTexture().bits());
			}else
				curbump=-1;

			this->gpuRT_->addPlane(Vector3f(plane->frame().position().v_), Vector4f(plane->getEquation()[0], plane->getEquation()[1], plane->getEquation()[2], plane->getEquation()[3]), 
				Vector3f(plane->material().diffuseColor().r, plane->material().diffuseColor().g, plane->material().diffuseColor().b),
				Vector3f(plane->material().specularColor().r, plane->material().specularColor().g, plane->material().specularColor().b) , 
				plane->material().specularCoefficient(), 
				curtex, curbump,
				Vector2f(1.0f/plane->material().textureScaleU(), 1.0f/plane->material().textureScaleV()),
				Vector3f(plane->material().reflectiveColor().r, plane->material().reflectiveColor().g, plane->material().reflectiveColor().b),
				Vector3f(plane->material().refractiveColor().r, plane->material().refractiveColor().g, plane->material().refractiveColor().b),
				1.0f/plane->material().refractiveIndex());

		}else if(cubeMap=dynamic_cast<CubeMap *>(this->scene()->getObject(i))){
			this->gpuRT_->addCubeMap(Vector3i(cubeMap->getImage(0).width(), cubeMap->getImage(0).width(), cubeMap->getImage(0).width()),  
				GL_BGRA, GL_UNSIGNED_BYTE, cubeMap->getImage(0).bits(), cubeMap->getImage(1).bits(), cubeMap->getImage(3).bits(), cubeMap->getImage(2).bits(), cubeMap->getImage(4).bits(), cubeMap->getImage(5).bits());

		}
	}


	//Signals connection
	/*for(int i=0; i<this->scene()->numLight(); i++){
		QObject::connect(&(this->scene()->getLight(i)->manipulatedFrame()), SIGNAL(manipulated(void)),
                         this, SLOT(updateGPURT(void)));
	}*/
	

}


/*void Viewer::updateGPURT(){
	//this->gpuRT_->updateScene();
}*/


#endif

void Viewer::print(const char *txt, float val){
	char buff[32];

	sprintf(buff, "%f", val);

	this->displayMessage(QString(txt) + QString(buff), 3000);
}